/*
   Be a text-based xload

   $dodge: tl.c,v 1.51 2007/06/18 20:20:20 dodge Exp $
   $Name: version_9 $
*/

#include "tl.h"

struct s_conf {
	int	use_lines;
	time_t	sleep_time;
	int	loadavg_to_use;
	int	all_loads;
};

// globals
WINDOW *win;
double		**loadavg_history;
unsigned int	  columns_last;
int		  tl_reinit_win;
struct s_conf	  conf;

// Display the right way to use this program. If we got here, it's not
// good. Just exit.
void
tl_usage (int argc, char **argv)
{
	fprintf(stderr, "tl\n");
	fprintf(stderr, "Version: %s\n", strlen(TL_VERSION_NAME) > 9 ?
			TL_VERSION_NAME : TL_VERSION_ID);
	fprintf(stderr, "Usage: %s [-s delay]\n", argv[0]);
	fprintf(stderr, "	-3		display three lines\n");
	fprintf(stderr, "	-d <filename>	send debugging output to filename\n");
	fprintf(stderr, "	-h		this stuff\n");
	fprintf(stderr, "	-l		use a line instead of a"
			" column\n");
	fprintf(stderr, "	-s delay	time between updates\n");
	exit(1);
}

// Get the current screen dimensions
void
get_geom(int *maxx, int *maxy)
{
	struct winsize w;

	debugout(__FILE__, __LINE__, "get_geom 00 -- %dx%d\n", maxx, maxy);

	ioctl(STDOUT_FILENO, TIOCGWINSZ, &w);
	*maxx = w.ws_col;
	*maxy = w.ws_row;
	resizeterm(*maxy, *maxx);
	wresize(win, *maxy, *maxx);
}

int
get_scale (int maxx, int maxy)
{
	int scale, x, i;

	debugout(__FILE__, __LINE__, "get_scale 00 -- %dx%d\n", maxx, maxy);
	scale = 1;

	for (i = 0; i < 3; i++)
	{
		for (x = 0; x < maxx; x++)
		{
			if (		loadavg_history != NULL &&
					loadavg_history[x] != NULL &&
					scale <= loadavg_history[x][i]
			   )
			{
				scale = loadavg_history[x][i] + 1;
			}
		}
	}

	return(scale);
}


void
draw_scale_lines(int maxx, int maxy, int scale, int rows_per_tick)
{
	int x, y;

	debugout(__FILE__, __LINE__, "draw_scale_lines 00 -- %dx%x -- "
			"scale:%d, rows_per_tick:%d\n", maxx, maxy, scale,
			rows_per_tick);

	for (x = 0; x < maxx; x++)
	{
		for (y = maxy - rows_per_tick; y > 0; y = y - rows_per_tick)
		{
			mvprintw(y, x, "_");
		}
	}
}


void
draw_line(int maxx, int maxy, int x, int scale, int rows_per_tick, int
		loadavg_to_use)
{
	int height, last_height;

	debugout(__FILE__, __LINE__, "draw_line 00 -- %dx%x -- x:%d, "
			"scale:%d, rows_per_tick:%d, loadavg_to_use:%d\n",
			maxx, maxy, x, scale, rows_per_tick,
			loadavg_to_use);

	height = -1;
	last_height = -1;

	if (rows_per_tick >= 4) {
		if (loadavg_history != NULL && loadavg_history[x] != NULL)
		{
			height = loadavg_history[x][loadavg_to_use] *
				rows_per_tick;
			if (x+1 < maxx && loadavg_history[x+1] != NULL)
			{

				last_height =
					loadavg_history[x+1][loadavg_to_use]
					* rows_per_tick;
			}
		}
	} else {
		if (loadavg_history != NULL && loadavg_history[x] != NULL)
		{
			height = (maxy - 1) * loadavg_history[x][loadavg_to_use] / scale;
			if (loadavg_history[x+1] != NULL)
			{
				last_height = (maxy - 1) *
					loadavg_history[x+1][loadavg_to_use]
					/ scale;
			}
		}
	}

	debugout(__FILE__, __LINE__, "draw_line 10 x:%d, height:%d\n", x,
			height);

	if (height == last_height)
	{
		mvprintw(maxy - height - 1, maxx -x -1, "-");

	} else if (height > last_height)
	{
		mvprintw(maxy - height - 1, maxx -x -1, "/");

	} else {
		mvprintw(maxy - height -1, maxx -x -1, "\\");
	}

}


void draw_bar (int maxx, int maxy, int x, int scale, int rows_per_tick,
		int loadavg_to_use)
{
	int y, height;

	debugout(__FILE__, __LINE__, "draw_bar 00 -- %dx%y, x:%d, "
			"scale:%d, raw_per_tick:%d, loadavg_to_use:%d\n",
			maxx, maxy, x, scale, rows_per_tick,
			loadavg_to_use);
	height = -1;

	if (loadavg_history == NULL || loadavg_history[x] == NULL)
	{
		return;
	}

	if (rows_per_tick >= 4) {
		height = loadavg_history[x][loadavg_to_use] * rows_per_tick;
	} else {
		height = maxy * loadavg_history[x][loadavg_to_use] / scale;
	}


	for (y = 0; y < height; y++)
	{
		mvprintw(maxy - y - 1, maxx - x - 1, "*");
	}
}

// draw the screen
void
draw_screen (int maxx, int maxy)
{
	int scale, rows_per_tick, x, i;
	char hostname[MAXHOSTNAMELEN];

	debugout(__FILE__, __LINE__, "draw_screen 00 -- %dx%x\n", maxx, maxy);
	get_geom(&maxx, &maxy);

	scale = get_scale(maxx, maxy);

	wclear(win);

	if (gethostname(hostname, MAXHOSTNAMELEN) == -1)
	{
		strncpy(hostname, "localhost", MAXHOSTNAMELEN);
	}

	rows_per_tick = (maxy - 1) / scale;

	mvprintw(0, 0, "%s:", hostname);
	for (i = 0; i < 3; i++)
	{
		if (conf.all_loads || i == conf.loadavg_to_use)
		{
			printw(" %0.2f", loadavg_history[0][i]);
		}
	}

	if (rows_per_tick < 4)
	{
		printw(" -- scale: %d", scale);
	}

	if (conf.use_lines)
	{
		if (rows_per_tick >= 4)
		{
			draw_scale_lines(maxx, maxy, scale, rows_per_tick);
		}
		for (i = 0; i < 3; i++)
		{
			if (conf.all_loads || i == conf.loadavg_to_use)
			{
				for (x = maxx - 1; x >= 0; x--)
				{
					draw_line(maxx, maxy, x, scale,
							rows_per_tick, i);
				}
			}
		}

	} else {
		for (x = maxx - 1; x >= 0; x--)
		{
			draw_bar(maxx, maxy, x, scale, rows_per_tick,
					conf.loadavg_to_use);
		}
		if (rows_per_tick >= 4)
		{
			draw_scale_lines(maxx, maxy, scale, rows_per_tick);
		}

	}

	mvwprintw(win, 0, maxx - 3, "%c",
			spin_chars[current_spin_char++]);
	if (current_spin_char >= strlen(spin_chars))
		current_spin_char = 0;

	wmove(win, 0, maxx - 2);

	refresh();
}

// the window has been resized. Do something about it here.
void
tl_resize (int sigraised)
{
	debugout(__FILE__, __LINE__, "tl_resize 00\n");
	tl_reinit_win = 1;
}

// Gets called when we get a signal. Cleanup the curses spooge and exit
// gracefully-ish
void
tl_cleanup (int sigraised)
{
	debugout(__FILE__, __LINE__, "tl_cleanup 00 -- sigraised:%d\n",
			sigraised);

	if (endwin() == ERR)
	{
		fprintf(stderr, "Can't endwin: %s\n", strerror(errno));
		exit(1);
	}

	if (sigraised != -1)
		fprintf(stderr, "Caught signal %d, cleaning up\n", sigraised);

	if (sigraised == SIGHUP || sigraised == SIGINT ||
		sigraised == SIGQUIT || sigraised == SIGKILL ||
		sigraised == SIGTERM || sigraised == -1)
	{
		exit(0);
	} else {
		abort();
	}
}

// miscellaneous initialization stuff. Go through the args, setup the
// screen.
void
tl_init (int argc, char **argv)
{
	int ch;

	loadavg_history = NULL;
	columns_last = 0;
	conf.sleep_time = 5;
	conf.use_lines = 0;
	spin_chars = SPIN_CHARS;
	current_spin_char = 0;

	while ((ch = getopt(argc, argv, "3d:h?ls:")) != -1)
	{
		switch (ch)
		{
			case '3':
				conf.all_loads = 1;
				conf.use_lines = 1;
				break;

			case 'd':
				debug_filename = (char *)optarg;
				break;

			case 'h':
				tl_usage(argc, argv);
				break;

			case '?':
				tl_usage(argc, argv);
				break;

			case 'l':
				conf.use_lines = 1;
				break;

			case 's':
				errno = 0;
				conf.sleep_time = strtol(optarg, NULL, 0);
				break;

			default:
				tl_usage(argc, argv);
		}
	}

	if (conf.sleep_time < 300)
	{
		conf.loadavg_to_use = 0;
	} else if (conf.sleep_time < 900)
	{
		conf.loadavg_to_use = 1;
	} else 
	{
		conf.loadavg_to_use = 2;
	}

	// Trap signals to cleanup as much as possible
	for (ch = 0; ch < 256; ch++)
	{
		if (ch != SIGABRT && ch != SIGWINCH)
		{
			signal(ch, &tl_cleanup);
		}
	}

	// trap the window-change signal
	signal(SIGWINCH, &tl_resize);

	// Setup the screen
	if ((win = initscr()) == NULL)
	{
		fprintf(stderr, "Can't initscr: %s\n", strerror(errno));
		exit(1);
	}
	cbreak();
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	halfdelay(10);
}


// Update the array of list of load averages. Most recent is at [0].
void
tl_add_load(int maxx, int maxy)
{
	double		**dummy,
			  loadavg[3];
	unsigned int	  columns_now;
	int		  i;

	if (columns_last > maxx)
	{
		columns_now = columns_last;
	} else {
		columns_now = maxx;
	}

	getloadavg(loadavg, 3);

	debugout(__FILE__, __LINE__, "tl_add_load 00 -- %dx%d -- " 
			"%0.2f %0.2f %0.2f\n", maxx, maxy, loadavg[0],
			loadavg[1], loadavg[2]);

	if (loadavg_history == NULL)
	{
		// Guess we've never been here. Get some memory
		loadavg_history = malloc(columns_now * sizeof(loadavg));
		debugout(__FILE__, __LINE__, "tl_add_load 10 -- "
				"loadavg_history is new, %d columns, "
				"at address %08x\n",
				columns_now, loadavg_history);

		if (loadavg_history == NULL)
		{
			fprintf(stderr, "Can't malloc %d bytes: %s\n",
					(int) (columns_now * sizeof(loadavg)),
					strerror(errno));
			tl_cleanup(-1);
		}

		for (i = 0; i < columns_now; i++)
		{
			loadavg_history[i] = NULL;
		}

		loadavg_history[0] = malloc(sizeof(loadavg));
		if (loadavg_history[0] == NULL)
		{
			fprintf(stderr, "Can't malloc %d bytes: %s\n",
					(int) sizeof(loadavg),
					strerror(errno));
			tl_cleanup(-1);
		}

		memcpy(loadavg_history[0], loadavg,
				sizeof(*loadavg_history));

		columns_last = columns_now;
		return;
	}

	if (columns_last != columns_now)
	{
		debugout(__FILE__, __LINE__,
				"tl_add_load 20 -- Number of columns changed from %d to %d\n",
				columns_last, columns_now);
		dummy = realloc(loadavg_history, sizeof(loadavg_history) * columns_now);
		if (dummy == NULL)
		{
			fprintf(stderr, "Can't malloc %d bytes: %s\n", (int)
					sizeof(loadavg_history) * columns_now,
					strerror(errno));
			tl_cleanup(-1);
		}

		if (loadavg_history != dummy)
		{
			debugout(__FILE__, __LINE__,
					"loadavg_history moving from %08x to %08x\n", 
					loadavg_history, dummy);
			loadavg_history = dummy;
		}

		for (i = columns_last; i < columns_now; i++)
		{
			loadavg_history[i] = NULL;
		}
	}

	for (i = columns_now - 1; i >= 1; i--)
	{
		loadavg_history[i] = loadavg_history[i -1];
	}

	loadavg_history[0] = malloc(sizeof(loadavg));
	if (loadavg_history[0] == NULL)
	{
		fprintf(stderr, "Can't malloc %d bytes: %s\n",
				(int) sizeof(loadavg), strerror(errno));
		tl_cleanup(-1);
	}
	memcpy(loadavg_history[0], loadavg, sizeof(loadavg));
	columns_last = columns_now;
}


int
main (int argc, char **argv)
{
	int maxy, maxx, retval;
	struct timeval now, refresh_time;

	tl_init(argc, argv);
	refresh_time.tv_sec = 0; 
	refresh_time.tv_usec = 0;
	debuginit();
	tl_reinit_win = 0;
	get_geom(&maxx, &maxy);
	while (1)
	{
		debugout(__FILE__, __LINE__, "main 10 -- %dx%d -- %d\n",
				maxx, maxy, tl_reinit_win);

		if (tl_reinit_win)
		{
			debugout(__FILE__, __LINE__, "main 11: %dx%d\n",
					maxx, maxy);
			tl_reinit_win = 0;
			endwin();
			win = initscr();
			get_geom(&maxx, &maxy);
			resizeterm(maxy, maxx);
			wresize(win, maxy, maxx);
			draw_screen(maxx, maxy);
		}

		tl_add_load(maxx, maxy);
		draw_screen(maxx, maxy);
		gettimeofday(&now, NULL);
		while (		now.tv_sec < refresh_time.tv_sec ||
				(now.tv_sec == refresh_time.tv_sec &&
				 now.tv_usec < refresh_time.tv_usec)
		      ) {
			debugout(__FILE__, __LINE__,
					"main 20 -- %d.%d < %d.%d\n",
					now.tv_sec, now.tv_usec,
					refresh_time.tv_sec,
					refresh_time.tv_usec);
			if ((retval = getch()) != ERR)
			{
				if ((char)retval == 'q')
				{
					tl_cleanup(-1);
				} else {
					draw_screen(maxx, maxy);
#ifdef DEBUG
					mvprintw(0, 35, "char:0x%x\n", retval);
#endif // DEBUG
					refresh();
				}
			}
			gettimeofday(&now, NULL);
		}
		refresh_time.tv_sec = now.tv_sec + conf.sleep_time;
		refresh_time.tv_usec = now.tv_usec;
	}
}
